from copy import copy
from pathlib import Path

import click
import pandas as pd
import yaml
from google.protobuf.timestamp_pb2 import Timestamp
from oaklib.implementations.pronto.pronto_implementation import ProntoImplementation
from oaklib.resource import OntologyResource
from phenopackets import (
    Diagnosis,
    Family,
    File,
    GeneDescriptor,
    GenomicInterpretation,
    Individual,
    Interpretation,
    MetaData,
    OntologyClass,
    Pedigree,
    Phenopacket,
    PhenotypicFeature,
    Resource,
    VariantInterpretation,
    VariationDescriptor,
    VcfRecord,
)
from pheval.prepare.create_noisy_phenopackets import load_ontology
from pheval.utils.file_utils import files_with_suffix
from pheval.utils.phenopacket_utils import create_hgnc_dict, write_phenopacket


def load_genotype_ontology():
    """Load genotype ontology"""
    genotype_resource = OntologyResource(slug="geno.owl", local=False)
    return ProntoImplementation(genotype_resource)


def exomiser_analysis_yml_reader(yaml_job_file_path: Path) -> dict:
    """Read an exomiser analysis yaml file."""
    with open(yaml_job_file_path) as yaml_job_file:
        yaml_job = yaml.safe_load(yaml_job_file)
    yaml_job_file.close()
    return yaml_job


def read_diagnoses_file(diagnoses_file_path: Path) -> pd.DataFrame:
    """Read a diagnoses file."""
    return pd.read_csv(diagnoses_file_path, delimiter="t")


def read_pedigree_file(pedigree_path: Path) -> list[str]:
    """Return the contents of a pedigree file"""
    return open(pedigree_path).readlines()


class ExomiserYamlToPhenopacketConverter:
    def __init__(self, genotype_ontology, human_phenotype_ontology, hgnc_data):
        self.genotype_ontology = genotype_ontology
        self.human_phenotype_ontology = human_phenotype_ontology
        self.hgnc_data = hgnc_data

    @staticmethod
    def construct_individual(yaml_job: dict, diagnoses: pd.DataFrame) -> Individual:
        """Construct individual for phenopacket."""
        return Individual(
            id=yaml_job["analysis"]["proband"],
            sex=diagnoses[diagnoses.ProbandId == yaml_job["analysis"]["proband"]]
            .iloc[0]["Sex"]
            .upper(),
        )

    @staticmethod
    def get_diagnoses_for_proband(yaml_job: dict, diagnoses: pd.DataFrame):
        """Get all diagnoses for proband."""
        return diagnoses.loc[diagnoses["ProbandId"] == yaml_job["analysis"]["proband"]]

    def construct_phenotypic_interpretations(self, yaml_job: dict) -> list[PhenotypicFeature]:
        """Construct the phenotypic features for the proband."""
        hpo_ids = yaml_job["analysis"]["hpoIds"]
        phenotypic_features = []
        for hpo_id in hpo_ids:
            try:
                rels = self.human_phenotype_ontology.entity_alias_map(hpo_id)
                hpo_term = "".join(rels[(list(rels.keys())[0])])
                hpo = PhenotypicFeature(type=OntologyClass(id=hpo_id, label=hpo_term))
                phenotypic_features.append(hpo)
            except AttributeError:
                hpo = PhenotypicFeature(type=OntologyClass(id=hpo_id))
                phenotypic_features.append(hpo)
        return phenotypic_features

    @staticmethod
    def construct_vcf_record(yaml_job: dict, diagnosis: pd.DataFrame) -> VcfRecord:
        """Construct the VCF record for a diagnosis."""
        return VcfRecord(
            genome_assembly=yaml_job["analysis"]["genomeAssembly"],
            chrom=diagnosis["Chr"],
            pos=int(diagnosis["Start"]),
            ref=str(diagnosis["Ref/Alt"]).split("/")[0],
            alt=str(diagnosis["Ref/Alt"]).split("/")[1],
        )

    def construct_allelic_state(self, diagnosis: pd.DataFrame) -> OntologyClass:
        """Construct the allelic state for a diagnosis."""
        return OntologyClass(
            id=list(self.genotype_ontology.basic_search(diagnosis["Genotype"].lower()))[0],
            label=diagnosis["Genotype"].lower(),
        )

    def construct_gene_descriptor(self, diagnosis: pd.DataFrame) -> GeneDescriptor:
        """Construct the Gene Descriptor for a diagnosis."""
        try:
            return GeneDescriptor(
                value_id=self.hgnc_data[diagnosis["Gene"]]["ensembl_id"],
                symbol=diagnosis["Gene"],
            )
        except KeyError:
            for _gene, gene_info in self.hgnc_data.items():
                for previous_name in gene_info["previous_names"]:
                    if diagnosis["Gene"] == previous_name:
                        return GeneDescriptor(
                            value_id=self.hgnc_data[gene_info["ensembl_id"]],
                            symbol=diagnosis["Gene"],
                        )

    def construct_variation_descriptor(
        self, yaml_job: dict, diagnosis: pd.DataFrame
    ) -> VariationDescriptor:
        """Construct a variation descriptor for a diagnosis."""
        return VariationDescriptor(
            id=yaml_job["analysis"]["proband"]
            + ":"
            + diagnosis["Chr"]
            + ":"
            + diagnosis["Start"]
            + ":"
            + diagnosis["Ref/Alt"],
            gene_context=self.construct_gene_descriptor(diagnosis),
            vcf_record=self.construct_vcf_record(yaml_job, diagnosis),
            allelic_state=self.construct_allelic_state(diagnosis),
        )

    def construct_variant_interpretation(
        self, yaml_job: dict, diagnosis: pd.DataFrame
    ) -> VariantInterpretation:
        """Construct the variant interpretation for a diagnosis."""
        return VariantInterpretation(
            variation_descriptor=self.construct_variation_descriptor(yaml_job, diagnosis),
        )

    def construct_genomic_interpretations(
        self, yaml_job: dict, diagnoses: pd.DataFrame
    ) -> list[GenomicInterpretation]:
        """Construct a list of genomic interpretations for a proband."""
        genomic_interpretations = []
        for _index, row in self.get_diagnoses_for_proband(yaml_job, diagnoses).iterrows():
            genomic_interpretation = GenomicInterpretation(
                subject_or_biosample_id=yaml_job["analysis"]["proband"],
                variant_interpretation=self.construct_variant_interpretation(
                    yaml_job=yaml_job, diagnosis=row
                ),
            )
            genomic_interpretations.append(genomic_interpretation)
        return genomic_interpretations

    def construct_diagnosis(self, yaml_job: dict, diagnoses: pd.DataFrame) -> Diagnosis:
        """Construct the diagnosis for a proband."""
        return Diagnosis(
            genomic_interpretations=self.construct_genomic_interpretations(yaml_job, diagnoses)
        )

    def construct_interpretations(
        self, yaml_job: dict, diagnoses: pd.DataFrame
    ) -> list[Interpretation]:
        """Construct interpretations for a proband."""
        return [
            Interpretation(
                id=yaml_job["analysis"]["proband"] + "-interpretation",
                diagnosis=self.construct_diagnosis(yaml_job, diagnoses),
            )
        ]

    @staticmethod
    def construct_meta_data() -> MetaData:
        """Construct the meta-data."""
        timestamp = Timestamp()
        timestamp.GetCurrentTime()
        return MetaData(
            created=timestamp,
            created_by="pheval-converter",
            resources=[
                Resource(
                    id="hp",
                    name="human phenotype ontology",
                    url="http://purl.obolibrary.org/obo/hp.owl",
                    version="hp/releases/2019-11-08",
                    namespace_prefix="HP",
                    iri_prefix="http://purl.obolibrary.org/obo/HP_",
                )
            ],
            phenopacket_schema_version="2.0",
        )

    @staticmethod
    def construct_files(yaml_job_file: dict) -> list[File]:
        """Construct the files."""
        return [
            File(
                uri=yaml_job_file["analysis"]["vcf"],
                file_attributes={
                    "fileFormat": "VCF",
                    "genomeAssembly": yaml_job_file["analysis"]["genomeAssembly"],
                },
            )
        ]


def construct_pedigree(pedigree: list[str]) -> tuple[str, Pedigree]:
    """Construct the pedigree message from a ped file."""
    persons = []
    family_id = None
    for individual in pedigree:
        entry = individual.split("\t")
        family_id = entry[0]
        sex = "."
        if (
            int(entry[4]) == 1
        ):  # until this is fixed with the phenopackets package, sex has to be reassigned
            sex = 2
        if int(entry[4]) == 2:
            sex = 1
        if str(entry[3]) == "0" and str(entry[2]) == "0":
            person = Pedigree.Person(
                family_id=family_id, individual_id=entry[1], sex=sex, affected_status=int(entry[5])
            )
            persons.append(person)
        if str(entry[3]) == "0" and str(entry[2]) != "0":
            person = Pedigree.Person(
                family_id=family_id,
                individual_id=entry[1],
                paternal_id=entry[2],
                sex=sex,
                affected_status=int(entry[5]),
            )
            persons.append(person)
        if str(entry[2]) == "0" and str(entry[3]) != "0":
            person = Pedigree.Person(
                family_id=family_id,
                individual_id=entry[1],
                maternal_id=entry[3],
                sex=sex,
                affected_status=int(entry[5]),
            )
            persons.append(person)
        if str(entry[2]) != "0" and str(entry[3] != "0"):
            person = Pedigree.Person(
                family_id=family_id,
                individual_id=entry[1],
                paternal_id=entry[2],
                maternal_id=entry[3],
                sex=sex,
                affected_status=int(entry[5]),
            )
            persons.append(person)
    return family_id, Pedigree(persons=persons)


def construct_phenopacket(
    yaml_job_file: dict,
    diagnoses: pd.DataFrame,
    exomiser_yaml_to_phenopacket_converter: ExomiserYamlToPhenopacketConverter,
) -> Phenopacket:
    """Construct a phenopacket."""
    return Phenopacket(
        id=yaml_job_file["analysis"]["proband"],
        subject=exomiser_yaml_to_phenopacket_converter.construct_individual(
            yaml_job=yaml_job_file, diagnoses=diagnoses
        ),
        phenotypic_features=exomiser_yaml_to_phenopacket_converter.construct_phenotypic_interpretations(
            yaml_job=yaml_job_file
        ),
        interpretations=exomiser_yaml_to_phenopacket_converter.construct_interpretations(
            yaml_job=yaml_job_file, diagnoses=diagnoses
        ),
        files=exomiser_yaml_to_phenopacket_converter.construct_files(yaml_job_file),
        meta_data=exomiser_yaml_to_phenopacket_converter.construct_meta_data(),
    )


def construct_family(
    yaml_job_file: dict,
    diagnoses: pd.DataFrame,
    exomiser_yaml_to_phenopacket_converter: ExomiserYamlToPhenopacketConverter,
    pedigree: list[str],
) -> Family:
    """Construct a Family"""
    phenopacket = construct_phenopacket(
        yaml_job_file, diagnoses, exomiser_yaml_to_phenopacket_converter
    )
    proband = copy(phenopacket)
    del proband.files[:]
    del proband.meta_data[:]
    family_id, ped = construct_pedigree(pedigree)
    return Family(
        id=family_id,
        proband=proband,
        pedigree=ped,
        files=phenopacket.files,
        meta_data=phenopacket.meta_data,
    )


def create_phenopacket(
    yaml_job_file: Path,
    diagnoses: pd.DataFrame,
    exomiser_converter: ExomiserYamlToPhenopacketConverter,
) -> Phenopacket or Family:
    """Construct either a family or phenopacket from an analysis yaml."""
    yaml_job = exomiser_analysis_yml_reader(yaml_job_file)
    phenopacket = (
        construct_phenopacket(yaml_job, diagnoses, exomiser_converter)
        if yaml_job["analysis"]["ped"] == ""
        else construct_family(
            yaml_job,
            diagnoses,
            exomiser_converter,
            read_pedigree_file(yaml_job["analysis"]["ped"]),
        )
    )
    return phenopacket


@click.command()
@click.option(
    "--directory",
    "-d",
    required=True,
    help="Directory for Exomiser yaml job files to be converted.",
    type=Path,
)
@click.option("--diagnoses-file", "-d", required=True, help="Diagnoses file", type=Path)
@click.option(
    "--output-dir", "-o", required=True, help="Output directory to write phenopackets", type=Path
)
def convert_exomiser_analysis_yamls_to_phenopacket(
    output_dir: Path, directory: Path, diagnoses_file: Path
):
    """Convert an Exomiser YAML file to a phenopacket schema given a .tsv diagnoses file containing the following
    required fields: ..."""
    try:
        output_dir.mkdir()
    except FileExistsError:
        pass
    diagnoses = read_diagnoses_file(diagnoses_file)
    exomiser_converter = ExomiserYamlToPhenopacketConverter(
        load_genotype_ontology(), load_ontology(), create_hgnc_dict()
    )
    for yaml_job_file in files_with_suffix(directory, ".yml"):
        phenopacket = create_phenopacket(yaml_job_file, diagnoses, exomiser_converter)
        write_phenopacket(
            phenopacket, output_dir.joinpath(yaml_job_file.name.replace(".yml", ".json"))
        )


@click.command()
@click.option(
    "--yaml-file",
    "-y",
    required=True,
    help="Path to Exomiser analysis yaml file for phenopacket conversion.",
    type=Path,
)
@click.option("--diagnoses-file", "-d", required=True, help="Diagnoses file", type=Path)
@click.option(
    "--output-dir", "-o", required=True, help="Output directory to write phenopackets", type=Path
)
def convert_exomiser_analysis_yaml_to_phenopacket(
    output_dir: Path, yaml_file: Path, diagnoses_file: Path
):
    """Convert Exomiser YAML files to the phenopacket schema given a .tsv diagnoses file containing the following
    required fields: ..."""
    try:
        output_dir.mkdir()
    except FileExistsError:
        pass
    diagnoses = read_diagnoses_file(diagnoses_file)
    exomiser_converter = ExomiserYamlToPhenopacketConverter(
        load_genotype_ontology(), load_ontology(), create_hgnc_dict()
    )
    phenopacket = create_phenopacket(yaml_file, diagnoses, exomiser_converter)
    write_phenopacket(phenopacket, Path(yaml_file.name + ".json"))



#!/usr/bin/python
import json
from pathlib import Path

import click
from pheval.post_processing.post_processing import (
    PhEvalDiseaseResult,
    PhEvalGeneResult,
    PhEvalVariantResult,
    generate_pheval_result,
)
from pheval.utils.file_utils import files_with_suffix


def read_exomiser_json_result(exomiser_result_path: Path) -> dict:
    """Load Exomiser json result."""
    with open(exomiser_result_path) as exomiser_json_result:
        exomiser_result = json.load(exomiser_json_result)
    exomiser_json_result.close()
    return exomiser_result


def trim_exomiser_result_filename(exomiser_result_path: Path) -> Path:
    """Trim suffix appended to Exomiser JSON result path."""
    return Path(str(exomiser_result_path.name).replace("-exomiser", ""))


class PhEvalGeneResultFromExomiserJsonCreator:
    def __init__(self, exomiser_json_result: [dict], score_name: str):
        self.exomiser_json_result = exomiser_json_result
        self.score_name = score_name

    @staticmethod
    def _find_gene_symbol(result_entry: dict) -> str:
        """Return gene symbol from Exomiser result entry."""
        return result_entry["geneSymbol"]

    @staticmethod
    def _find_gene_identifier(result_entry: dict) -> str:
        """Return ensembl gene identifier from Exomiser result entry."""
        return result_entry["geneIdentifier"]["geneId"]

    def _find_relevant_score(self, result_entry: dict):
        """Return score from Exomiser result entry."""
        return round(result_entry[self.score_name], 4)

    def extract_pheval_gene_requirements(self) -> [PhEvalGeneResult]:
        """Extract data required to produce PhEval gene output."""
        simplified_exomiser_result = []
        for result_entry in self.exomiser_json_result:
            if self.score_name in result_entry:
                simplified_exomiser_result.append(
                    PhEvalGeneResult(
                        gene_symbol=self._find_gene_symbol(result_entry),
                        gene_identifier=self._find_gene_identifier(result_entry),
                        score=self._find_relevant_score(result_entry),
                    )
                )

        return simplified_exomiser_result


class PhEvalVariantResultFromExomiserJsonCreator:

    def __init__(self, exomiser_json_result: [dict], score_name: str):
        self.exomiser_json_result = exomiser_json_result
        self.score_name = score_name

    @staticmethod
    def _find_chromosome(result_entry: dict) -> str:
        """Return chromosome from Exomiser result entry."""
        return result_entry["contigName"]

    @staticmethod
    def _find_start_pos(result_entry: dict) -> int:
        """Return start position from Exomiser result entry."""
        return result_entry["start"]

    @staticmethod
    def _find_end_pos(result_entry: dict) -> int:
        """Return end position from Exomiser result entry."""
        return result_entry["end"]

    @staticmethod
    def _find_ref(result_entry: dict) -> str:
        """Return reference allele from Exomiser result entry."""
        return result_entry["ref"]

    @staticmethod
    def _find_alt(result_entry: dict) -> str:
        """Return alternate allele from Exomiser result entry."""
        if "alt" in result_entry and result_entry["alt"] is not None:
            return result_entry["alt"].strip(">").strip("<")
        else:
            return ""

    def _find_relevant_score(self, result_entry) -> float:
        """Return score from Exomiser result entry."""
        return round(result_entry[self.score_name], 4)

    def _filter_for_acmg_assignments(
        self, variant: PhEvalVariantResult, score: float, variant_acmg_assignments: dict
    ) -> bool:
        """Filter variants if they meet the PATHOGENIC or LIKELY_PATHOGENIC ACMG classification."""
        for assignment in variant_acmg_assignments:
            if variant == PhEvalVariantResult(
                chromosome=self._find_chromosome(assignment["variantEvaluation"]),
                start=self._find_start_pos(assignment["variantEvaluation"]),
                end=self._find_end_pos(assignment["variantEvaluation"]),
                ref=self._find_ref(assignment["variantEvaluation"]),
                alt=self._find_alt(assignment["variantEvaluation"]),
                score=score,
            ) and (
                assignment["acmgClassification"] == "PATHOGENIC"
                or assignment["acmgClassification"] == "LIKELY_PATHOGENIC"
            ):
                return True

    def extract_pheval_variant_requirements(
        self, use_acmg_filter: bool = False
    ) -> [PhEvalVariantResult]:
        """Extract data required to produce PhEval variant output."""
        simplified_exomiser_result = []
        for result_entry in self.exomiser_json_result:
            for gene_hit in result_entry["geneScores"]:
                if self.score_name in result_entry:
                    if "contributingVariants" in gene_hit:
                        score = self._find_relevant_score(result_entry)
                        contributing_variants = gene_hit["contributingVariants"]
                        variant_acmg_assignments = gene_hit["acmgAssignments"]
                        for cv in contributing_variants:
                            variant = PhEvalVariantResult(
                                chromosome=self._find_chromosome(cv),
                                start=self._find_start_pos(cv),
                                end=self._find_end_pos(cv),
                                ref=self._find_ref(cv),
                                alt=self._find_alt(cv),
                                score=score,
                            )
                            if use_acmg_filter and self._filter_for_acmg_assignments(
                                variant, score, variant_acmg_assignments
                            ):
                                simplified_exomiser_result.append(variant)
                            if not use_acmg_filter:
                                simplified_exomiser_result.append(variant)
        return simplified_exomiser_result


class PhEvalDiseaseResultFromExomiserJsonCreator:
    def __init__(self, exomiser_json_result: [dict]):
        self.exomiser_json_result = exomiser_json_result

    @staticmethod
    def _find_disease_name(result_entry: dict) -> str:
        """Return disease term from Exomiser result entry."""
        return result_entry["diseaseTerm"]

    @staticmethod
    def _find_disease_identifier(result_entry: dict) -> int:
        """Return disease ID from Exomiser result entry."""
        return result_entry["diseaseId"]

    @staticmethod
    def _find_relevant_score(result_entry) -> float:
        """Return score from Exomiser result entry."""
        return round(result_entry["score"], 4)

    def extract_pheval_disease_requirements(self) -> [PhEvalDiseaseResult]:
        """Extract data required to produce PhEval disease output."""
        simplified_exomiser_result = []
        for result_entry in self.exomiser_json_result:
            try:
                for disease in result_entry["priorityResults"]["HIPHIVE_PRIORITY"][
                    "diseaseMatches"
                ]:
                    simplified_exomiser_result.append(
                        PhEvalDiseaseResult(
                            disease_name=self._find_disease_name(disease["model"]),
                            disease_identifier=self._find_disease_identifier(disease["model"]),
                            score=self._find_relevant_score(disease),
                        )
                    )
            except KeyError:
                pass
        return simplified_exomiser_result


def create_standardised_results(
    results_dir: Path,
    output_dir: Path,
    score_name: str,
    sort_order: str,
    variant_analysis: bool,
    gene_analysis: bool,
    disease_analysis: bool,
    include_acmg: bool = False,
) -> None:
    """Write standardised gene/variant/disease results from default Exomiser json output."""
    for exomiser_json_result in files_with_suffix(results_dir, ".json"):
        exomiser_result = read_exomiser_json_result(exomiser_json_result)
        if gene_analysis:
            pheval_gene_requirements = PhEvalGeneResultFromExomiserJsonCreator(
                exomiser_result, score_name
            ).extract_pheval_gene_requirements()
            generate_pheval_result(
                pheval_result=pheval_gene_requirements,
                sort_order_str=sort_order,
                output_dir=output_dir,
                tool_result_path=trim_exomiser_result_filename(exomiser_json_result),
            )
        if variant_analysis:
            pheval_variant_requirements = PhEvalVariantResultFromExomiserJsonCreator(
                exomiser_result, score_name
            ).extract_pheval_variant_requirements(include_acmg)
            generate_pheval_result(
                pheval_result=pheval_variant_requirements,
                sort_order_str=sort_order,
                output_dir=output_dir,
                tool_result_path=trim_exomiser_result_filename(exomiser_json_result),
            )
        if disease_analysis:
            pheval_disease_requirements = PhEvalDiseaseResultFromExomiserJsonCreator(
                exomiser_result
            ).extract_pheval_disease_requirements()
            generate_pheval_result(
                pheval_result=pheval_disease_requirements,
                sort_order_str=sort_order,
                output_dir=output_dir,
                tool_result_path=trim_exomiser_result_filename(exomiser_json_result),
            )


@click.command()
@click.option(
    "--output-dir",
    "-o",
    required=True,
    metavar="PATH",
    help="Output directory for standardised results.",
    type=Path,
)
@click.option(
    "--results-dir",
    "-R",
    required=True,
    metavar="DIRECTORY",
    help="Full path to Exomiser results directory to be standardised.",
    type=Path,
)
@click.option(
    "--score-name",
    "-s",
    required=True,
    help="Score name to extract from results.",
    type=click.Choice(["combinedScore", "priorityScore", "variantScore", "pValue"]),
    default="combinedScore",
    show_default=True,
)
@click.option(
    "--sort-order",
    "-so",
    required=True,
    help="Ordering of results for ranking.",
    type=click.Choice(["ascending", "descending"]),
    default="descending",
    show_default=True,
)
@click.option(
    "--gene-analysis/--no-gene-analysis",
    type=bool,
    default=False,
    help="Specify whether to create PhEval gene results.",
)
@click.option(
    "--variant-analysis/--no-variant-analysis",
    type=bool,
    default=False,
    help="Specify whether to create PhEval variant results.",
)
@click.option(
    "--disease-analysis/--no-disease-analysis",
    type=bool,
    default=False,
    help="Specify whether to create PhEval disease results.",
)
@click.option(
    "--include-acmg",
    is_flag=True,
    type=bool,
    default=False,
    help="Specify whether to include ACMG filter for PATHOGENIC or LIKELY_PATHOGENIC classifications.",
)
def post_process_exomiser_results(
    output_dir: Path,
    results_dir: Path,
    score_name: str,
    sort_order: str,
    gene_analysis: bool,
    variant_analysis: bool,
    disease_analysis: bool,
    include_acmg: bool,
):
    """Post-process Exomiser json results into PhEval gene and variant outputs."""
    (
        output_dir.joinpath("pheval_gene_results").mkdir(parents=True, exist_ok=True)
        if gene_analysis
        else None
    )
    (
        output_dir.joinpath("pheval_variant_results").mkdir(parents=True, exist_ok=True)
        if variant_analysis
        else None
    )
    (
        output_dir.joinpath("pheval_disease_results").mkdir(parents=True, exist_ok=True)
        if disease_analysis
        else None
    )
    create_standardised_results(
        results_dir,
        output_dir,
        score_name,
        sort_order,
        variant_analysis,
        gene_analysis,
        disease_analysis,
        include_acmg,
    )
